#include <cstdio>
#include "input.h"

char readChar() {
	char r;
	scanf(" %c", &r);
	return r;
}

int readInt() {
	int r;
	scanf("%d", &r);
	return r;
}

float readFloat() {
	float r;
	scanf("%f", &r);
	return r;
}
