#pragma once

/**
 * This file contains instance-level configuration that is set up by parsing the command line.
 */

// Debug level.
extern int debugLevel;

// Force full re-compilation.
extern bool force;
